import 'package:flutter/material.dart';
import 'package:flutter_custom_toolbox/flutter_toolbox.dart';

import 'package:hangman/data/fetch_data_helper.dart';
import 'package:hangman/models/game/game.dart';
import 'package:hangman/models/game/picked_word.dart';
import 'package:hangman/models/settings/settings_game.dart';
import 'package:hangman/models/settings/settings_global.dart';

part 'game_state.dart';

class GameCubit extends HydratedCubit<GameState> {
  GameCubit()
      : super(GameState(
          currentGame: Game.createNull(),
        ));

  void updateState(Game game) {
    emit(GameState(
      currentGame: game,
    ));
  }

  void refresh() {
    final Game game = Game(
      // Settings
      gameSettings: state.currentGame.gameSettings,
      globalSettings: state.currentGame.globalSettings,
      // State
      isRunning: state.currentGame.isRunning,
      isStarted: state.currentGame.isStarted,
      isFinished: state.currentGame.isFinished,
      animationInProgress: state.currentGame.animationInProgress,
      gettingSecretWord: state.currentGame.gettingSecretWord,
      // Base data
      secretWord: state.currentGame.secretWord,
      clue: state.currentGame.clue,
      // Game data
      hiddenWord: state.currentGame.hiddenWord,
      usedLetters: state.currentGame.usedLetters,
      errorsCount: state.currentGame.errorsCount,
      isClueDisplayed: state.currentGame.isClueDisplayed,
    );
    // game.dump();

    updateState(game);
  }

  void startNewGame({
    required GameSettings gameSettings,
    required GlobalSettings globalSettings,
  }) {
    final Game newGame = Game.createNew(
      // Settings
      gameSettings: gameSettings,
      globalSettings: globalSettings,
    );

    updateState(newGame);
    refresh();

    FetchDataHelper().pickRandomWord(gameSettings).then((PickedWord pickedWord) {
      state.currentGame.gettingSecretWord = false;
      state.currentGame.secretWord = pickedWord.word;
      state.currentGame.clue = pickedWord.clue;
      state.currentGame.hiddenWord = List<String>.generate(pickedWord.word.length, (i) => '_');
      state.currentGame.dump();

      refresh();
    });
  }

  void quitGame() {
    state.currentGame.isRunning = false;
    refresh();
  }

  void resumeSavedGame() {
    state.currentGame.isRunning = true;
    refresh();
  }

  void deleteSavedGame() {
    state.currentGame.isRunning = false;
    state.currentGame.isFinished = true;
    refresh();
  }

  void submitLetter(String key) {
    if (!state.currentGame.usedLetters.contains(key)) {
      updateUsedLetters(key);

      if (state.currentGame.secretWord.contains(key)) {
        for (int index = 0; index < state.currentGame.secretWord.length; index++) {
          if (key == state.currentGame.secretWord[index]) {
            updateHiddenWord(index, key);
          }
        }
      } else {
        incrementErrorsCount();
      }
    }

    if ((state.currentGame.hiddenWordAsString == state.currentGame.secretWord) ||
        (state.currentGame.errorsCount == 8)) {
      state.currentGame.isFinished = true;
    }

    refresh();
  }

  void updateUsedLetters(String key) {
    state.currentGame.usedLetters.add(key);
    refresh();
  }

  void updateHiddenWord(int index, String letter) {
    state.currentGame.hiddenWord[index] = letter;
    refresh();
  }

  void incrementErrorsCount() {
    state.currentGame.errorsCount++;
    refresh();
  }

  void toggleClueDisplay() {
    state.currentGame.isClueDisplayed = !state.currentGame.isClueDisplayed;
    refresh();
  }

  @override
  GameState? fromJson(Map<String, dynamic> json) {
    final Game currentGame = json['currentGame'] as Game;

    return GameState(
      currentGame: currentGame,
    );
  }

  @override
  Map<String, dynamic>? toJson(GameState state) {
    return <String, dynamic>{
      'currentGame': state.currentGame.toJson(),
    };
  }
}
